#include "license_query.h"
#include "license.h"
#include <lua.h>
#include <lauxlib.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#include "license_objects.h"
#include "license_logging.h"
#include "license_lua_userdata.h"
#include "license_env.h"
#include "string_list.h"
#include "metric_list.h"
#include "xalloc.h"

struct op5lic_query_t {
	struct op5lic_t *lic;
	char *action;
	struct op5lic_env_list_t * environment;
};

struct op5lic_query_response_t {
	bool allow;
	struct string_list_t *message_list;
	struct metric_list_t *metric_list;
};

struct op5lic_query_t *op5lic_query_create(struct op5lic_t *lic,
		const char *action, struct op5lic_env_list_t *environment) {
	struct op5lic_query_t *lq;
	if (!lic | !action)
		return NULL;

	lq = xcalloc(1, sizeof(struct op5lic_query_t));
	lq->lic = lic;
	lq->action = xstrdup(action);
	lq->environment = environment;
	return lq;
}

struct op5lic_query_response_t *op5lic_query_response_new(void) {
	struct op5lic_query_response_t *qr = xcalloc(1, sizeof(struct op5lic_query_response_t));
	return qr;
}

static void op5lic_query_response_set_message_list(struct op5lic_query_response_t *qresponse, const struct string_list_t *messages) {
	qresponse->message_list = string_list_copy(messages);
}

static void op5lic_query_response_set_metric_list(struct op5lic_query_response_t *qresponse, const struct metric_list_t *metrics) {
	qresponse->metric_list = metric_list_copy(metrics);
}

static void op5lic_query_response_set_allow(struct op5lic_query_response_t *qresponse, bool allow) {
	qresponse->allow = allow;
}

struct string_list_t * op5lic_query_response_message_list(const struct op5lic_query_response_t *qresponse) {
	if (!qresponse)
		return NULL;

	return qresponse->message_list;
}

struct metric_list_t * op5lic_query_response_metric_list(const struct op5lic_query_response_t *qresponse) {
	if (!qresponse)
		return NULL;

	return qresponse->metric_list;
}


bool op5lic_query_response_allow(const struct op5lic_query_response_t *qresponse) {
	return qresponse->allow;
}

void op5lic_query_response_destroy(struct op5lic_query_response_t *qresponse) {
	string_list_destroy(qresponse->message_list);
	metric_list_destroy(qresponse->metric_list);
	xfree(qresponse);
}
void op5lic_query_destroy(struct op5lic_query_t *lq) {
	xfree(lq->action);
	xfree(lq);
}

/*******************************************************************************
 * Query execution
 ******************************************************************************/
struct op5lic_query_response_t * op5lic_query_run_with_entry_point(struct op5lic_query_t *lq, struct op5lic_error_t *err, const char *entry_point) {
	int status;
	struct op5lic_query_response_udata_t *qresponse;
	char *envstr;

	/* Fetch function */
	lua_getglobal(lq->lic->luaState, entry_point);
	if (!lua_isfunction(lq->lic->luaState, -1)) {
		log_warn("Expected function `%s()' to be defined, but it's not.", entry_point);
		op5lic_error_set(err, OP5LIC_STATUS_INVALID_ENTRYPOINT, entry_point);
		return NULL;
	}
	qresponse = op5lic_query_response_udata_new(lq->lic->luaState);
	lua_pushstring(lq->lic->luaState, lq->action);
	op5lic_query_env_udata_new(lq->lic->luaState, lq->environment);

	log_debug("Executing query for action `%s()', entry point is `%s()'", lq->action, entry_point);
	envstr = op5lic_env_list_to_string(lq->environment);
	log_debug("Supplied environment is: %s", envstr);
	xfree(envstr);

	status = lua_pcall(lq->lic->luaState, 3, 1, 0);
	bool allow = false;
	if (status != 0) {
		log_warn("Failed to run method: `%s'",
				lua_tostring(lq->lic->luaState, -1));
		op5lic_error_set(err, op5lic_error_from_luastatus(status),
				lua_tostring(lq->lic->luaState, -1));
		lua_pop(lq->lic->luaState, 1);
		return NULL;
	}
	allow = lua_toboolean(lq->lic->luaState, -1);
	lua_pop(lq->lic->luaState, 1);
	log_debug("License %s action %s", (allow ? "allows" : "disallows"), lq->action);
	/* TODO: We create a new, separate object which is pretty much identical to the
	 * udata one, here. Except that this one isn't garbage collected by Lua, and it has its own
	 * related functions.
	 *
	 * Another option would be to keep a reference to our udata on the stack to
	 * keep it from being GC'd. If we do that, however, we should provide
	 * aliases that do not expose the fact that the (API) user is dealing with
	 * something coming from Lua-land, for tidyness win.
	 *
	 * Still haven't decided which I dislike the least.
	 * */
	struct op5lic_query_response_t * response = op5lic_query_response_new();
	op5lic_query_response_set_message_list(
			response,
			qresponse->message_list
			);

	op5lic_query_response_set_metric_list(
			response,
			qresponse->metric_list
			);

	op5lic_query_response_set_allow(
			response,
			allow
			);
	return response;
}

/* Load table at -1 into metadata md */
void load_table(lua_State *L, struct op5lic_metadata_t *md) {
	const char *fieldname;
	int ftype;
	lua_pushnil(L); /*first key*/
	while (lua_next(L, -2) != 0) {
		struct op5lic_metadata_t *tmp;
		fieldname = luaL_checkstring(L, -2);
		ftype = lua_type(L, -1);

		log_debug("%s => %s",
				lua_typename(L, lua_type(L, -2)),
				lua_typename(L, ftype));
		switch (ftype) {
		case LUA_TTABLE:
			tmp = op5lic_metadata_create();
			load_table(L, tmp);
			op5lic_metadata_add_field_table(md, fieldname, tmp);
			break;
		case LUA_TSTRING:
			op5lic_metadata_add_field_str(md, fieldname, luaL_checkstring(L, -1));
			break;
		default:
			log_err("Unhandled metadata value type %s on key %s ... Skipping.", lua_typename(L, ftype), fieldname);
			break;
		}
		lua_pop(L, 1);
	}
}

struct op5lic_metadata_t * op5lic_query_metadata(struct op5lic_t *lic, struct op5lic_error_t *err) {
	struct op5lic_metadata_t *md;
	lua_State *L = lic->luaState;

	lua_getglobal(L, "metadata");
	if (!lua_istable(L, -1)) {
		log_warn("Expected table `metadata' to be defined, but it's not.");
		op5lic_error_set(err, OP5LIC_STATUS_MISSING_METADATA, NULL);
		return NULL;
	}

	md = op5lic_metadata_create();
	load_table(L, md);
	return md;
}
struct op5lic_query_response_t * op5lic_query_run(struct op5lic_query_t *lq, struct op5lic_error_t *err) {
	return op5lic_query_run_with_entry_point(lq, err, OP5LIC_DEFAULT_ENTRY_POINT_NAME);
}
